"""
Credits:

* `gist@ninely <https://gist.github.com/ninely/88485b2e265d852d3feb8bd115065b1a>`_
* `langchain@#1705 <https://github.com/hwchase17/langchain/discussions/1706>`_
"""
from typing import Any, Awaitable, Callable, Optional, Union

from fastapi.responses import StreamingResponse as _StreamingResponse
from langchain.chains.base import Chain
from starlette.background import BackgroundTask
from starlette.types import Send

from lanarky.callbacks import get_streaming_callback


class StreamingResponse(_StreamingResponse):
    """StreamingResponse class wrapper for langchain chains."""

    def __init__(
        self,
        chain_executor: Callable[[Send], Awaitable[Any]],
        background: Optional[BackgroundTask] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(content=iter(()), background=background, **kwargs)

        self.chain_executor = chain_executor

    async def stream_response(self, send: Send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )

        try:
            outputs = await self.chain_executor(send)
            if self.background is not None:
                self.background.kwargs["outputs"] = outputs
        except Exception as e:
            if self.background is not None:
                self.background.kwargs["outputs"] = str(e)
            await send(
                {
                    "type": "http.response.body",
                    "body": str(e).encode(self.charset),
                    "more_body": False,
                }
            )
            return

        await send({"type": "http.response.body", "body": b"", "more_body": False})

    @staticmethod
    def _create_chain_executor(
        chain: Chain, inputs: Union[dict[str, Any], Any], **callback_kwargs
    ) -> Callable[[Send], Awaitable[Any]]:
        async def wrapper(send: Send):
            return await chain.acall(
                inputs=inputs,
                callbacks=[get_streaming_callback(chain, send=send, **callback_kwargs)],
            )

        return wrapper

    @classmethod
    def from_chain(
        cls,
        chain: Chain,
        inputs: Union[dict[str, Any], Any],
        background: Optional[BackgroundTask] = None,
        callback_kwargs: dict[str, Any] = {},
        **kwargs: Any,
    ) -> "StreamingResponse":
        chain_executor = cls._create_chain_executor(chain, inputs, **callback_kwargs)

        return cls(
            chain_executor=chain_executor,
            background=background,
            **kwargs,
        )



from langchain.chains.base import Chain

from lanarky.register import STREAMING_CALLBACKS, WEBSOCKET_CALLBACKS

from .agents import AsyncAgentsStreamingCallback, AsyncAgentsWebsocketCallback
from .base import AsyncStreamingResponseCallback, AsyncWebsocketCallback
from .llm import (
    AsyncConversationChainStreamingCallback,
    AsyncConversationChainWebsocketCallback,
    AsyncLLMChainStreamingCallback,
    AsyncLLMChainWebsocketCallback,
)
from .qa_with_sources import (
    AsyncConversationalRetrievalChainStreamingCallback,
    AsyncConversationalRetrievalChainWebsocketCallback,
    AsyncQAWithSourcesChainStreamingCallback,
    AsyncQAWithSourcesChainWebsocketCallback,
    AsyncRetrievalQAWithSourcesChainStreamingCallback,
    AsyncRetrievalQAWithSourcesChainWebsocketCallback,
    AsyncVectorDBQAWithSourcesChainStreamingCallback,
    AsyncVectorDBQAWithSourcesChainWebsocketCallback,
)
from .retrieval_qa import (
    AsyncRetrievalQAStreamingCallback,
    AsyncRetrievalQAWebsocketCallback,
    AsyncVectorDBQAStreamingCallback,
    AsyncVectorDBQAWebsocketCallback,
)

__all__ = [
    "AsyncLLMChainStreamingCallback",
    "AsyncLLMChainWebsocketCallback",
    "AsyncConversationChainStreamingCallback",
    "AsyncConversationChainWebsocketCallback",
    "AsyncRetrievalQAStreamingCallback",
    "AsyncRetrievalQAWebsocketCallback",
    "AsyncVectorDBQAStreamingCallback",
    "AsyncVectorDBQAWebsocketCallback",
    "AsyncQAWithSourcesChainStreamingCallback",
    "AsyncQAWithSourcesChainWebsocketCallback",
    "AsyncRetrievalQAWithSourcesChainStreamingCallback",
    "AsyncRetrievalQAWithSourcesChainWebsocketCallback",
    "AsyncVectorDBQAWithSourcesChainStreamingCallback",
    "AsyncVectorDBQAWithSourcesChainWebsocketCallback",
    "AsyncConversationalRetrievalChainStreamingCallback",
    "AsyncConversationalRetrievalChainWebsocketCallback",
    "AsyncAgentsStreamingCallback",
    "AsyncAgentsWebsocketCallback",
]

ERROR_MESSAGE = """Error! Chain type '{chain_type}' is not currently supported by '{callable_name}'."""


def get_streaming_callback(
    chain: Chain, *args, **kwargs
) -> AsyncStreamingResponseCallback:
    """Get the streaming callback for the given chain type."""
    chain_type = chain.__class__.__name__
    try:
        callback = STREAMING_CALLBACKS[chain_type]
        return callback(*args, **kwargs)
    except KeyError:
        raise KeyError(
            ERROR_MESSAGE.format(
                chain_type=chain_type, callable_name="AsyncStreamingResponseCallback"
            )
        )


def get_websocket_callback(chain: Chain, *args, **kwargs) -> AsyncWebsocketCallback:
    """Get the websocket callback for the given chain type."""
    chain_type = chain.__class__.__name__
    try:
        callback = WEBSOCKET_CALLBACKS[chain_type]
        return callback(*args, **kwargs)
    except KeyError:
        raise KeyError(
            ERROR_MESSAGE.format(
                chain_type=chain_type, callable_name="AsyncWebsocketCallback"
            )
        )



from lanarky.register import register_streaming_callback, register_websocket_callback

from .retrieval_qa import (
    AsyncBaseRetrievalQAStreamingCallback,
    AsyncBaseRetrievalQAWebsocketCallback,
)


@register_streaming_callback("BaseQAWithSources")
class AsyncBaseQAWithSourcesChainStreamingCallback(
    AsyncBaseRetrievalQAStreamingCallback
):
    """AsyncStreamingResponseCallback handler for BaseQAWithSources."""

    pass


@register_websocket_callback("BaseQAWithSources")
class AsyncBaseQAWithSourcesChainWebsocketCallback(
    AsyncBaseRetrievalQAWebsocketCallback
):
    """AsyncWebsocketCallback handler for BaseQAWithSources."""

    pass


@register_streaming_callback("QAWithSourcesChain")
class AsyncQAWithSourcesChainStreamingCallback(
    AsyncBaseQAWithSourcesChainStreamingCallback
):
    """AsyncStreamingResponseCallback handler for QAWithSourcesChain."""

    pass


@register_websocket_callback("QAWithSourcesChain")
class AsyncQAWithSourcesChainWebsocketCallback(
    AsyncBaseQAWithSourcesChainWebsocketCallback
):
    """AsyncWebsocketCallback handler for QAWithSourcesChain."""

    pass


@register_streaming_callback("VectorDBQAWithSourcesChain")
class AsyncVectorDBQAWithSourcesChainStreamingCallback(
    AsyncBaseQAWithSourcesChainStreamingCallback
):
    """AsyncStreamingResponseCallback handler for VectorDBQAWithSourcesChain."""

    pass


@register_websocket_callback("VectorDBQAWithSourcesChain")
class AsyncVectorDBQAWithSourcesChainWebsocketCallback(
    AsyncBaseQAWithSourcesChainWebsocketCallback
):
    """AsyncWebsocketCallback handler for VectorDBQAWithSourcesChain."""

    pass


@register_streaming_callback("RetrievalQAWithSourcesChain")
class AsyncRetrievalQAWithSourcesChainStreamingCallback(
    AsyncBaseQAWithSourcesChainStreamingCallback
):
    """AsyncStreamingResponseCallback handler for RetrievalQAWithSourcesChain."""

    pass


@register_websocket_callback("RetrievalQAWithSourcesChain")
class AsyncRetrievalQAWithSourcesChainWebsocketCallback(
    AsyncBaseQAWithSourcesChainWebsocketCallback
):
    """AsyncWebsocketCallback handler for RetrievalQAWithSourcesChain."""

    pass


@register_streaming_callback("ConversationalRetrievalChain")
class AsyncConversationalRetrievalChainStreamingCallback(
    AsyncBaseQAWithSourcesChainStreamingCallback
):
    """AsyncStreamingResponseCallback handler for ConversationalRetrievalChain."""

    pass


@register_websocket_callback("ConversationalRetrievalChain")
class AsyncConversationalRetrievalChainWebsocketCallback(
    AsyncBaseQAWithSourcesChainWebsocketCallback
):
    """AsyncWebsocketCallback handler for ConversationalRetrievalChain."""

    pass



from typing import Any

from lanarky.register import register_streaming_callback, register_websocket_callback

from .base import AsyncStreamingResponseCallback, AsyncWebsocketCallback


@register_streaming_callback("LLMChain")
class AsyncLLMChainStreamingCallback(AsyncStreamingResponseCallback):
    """AsyncStreamingResponseCallback handler for LLMChain."""

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        message = self._construct_message(token)
        await self.send(message)


@register_websocket_callback("LLMChain")
class AsyncLLMChainWebsocketCallback(AsyncWebsocketCallback):
    """AsyncWebsocketCallback handler for LLMChain."""

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        message = self._construct_message(token)
        await self.websocket.send_json(message)


@register_streaming_callback("ConversationChain")
class AsyncConversationChainStreamingCallback(AsyncLLMChainStreamingCallback):
    """AsyncStreamingResponseCallback handler for ConversationChain."""

    pass


@register_websocket_callback("ConversationChain")
class AsyncConversationChainWebsocketCallback(AsyncLLMChainWebsocketCallback):
    """AsyncWebsocketCallback handler for ConversationChain."""

    pass



from typing import Any, Dict

from lanarky.register import register_streaming_callback, register_websocket_callback

from .llm import AsyncLLMChainStreamingCallback, AsyncLLMChainWebsocketCallback

SOURCE_DOCUMENT_TEMPLATE = """
page content: {page_content}
source: {source}
"""


class AsyncBaseRetrievalQAStreamingCallback(AsyncLLMChainStreamingCallback):
    """AsyncStreamingResponseCallback handler for BaseRetrievalQA."""

    source_document_template: str = SOURCE_DOCUMENT_TEMPLATE

    async def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Run when chain ends running."""
        if "source_documents" in outputs:
            message = self._construct_message("\n\nSOURCE DOCUMENTS: \n")
            await self.send(message)
            for document in outputs["source_documents"]:
                message = self._construct_message(
                    self.source_document_template.format(
                        page_content=document.page_content,
                        source=document.metadata["source"],
                    )
                )
                await self.send(message)


class AsyncBaseRetrievalQAWebsocketCallback(AsyncLLMChainWebsocketCallback):
    """AsyncWebsocketCallback handler for BaseRetrievalQA."""

    source_document_template: str = SOURCE_DOCUMENT_TEMPLATE

    async def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Run when chain ends running."""
        if "source_documents" in outputs:
            message = self._construct_message("\n\nSOURCE DOCUMENTS: \n")
            await self.websocket.send_json(message)
            for document in outputs["source_documents"]:
                message = self._construct_message(
                    self.source_document_template.format(
                        page_content=document.page_content,
                        source=document.metadata["source"],
                    )
                )
                await self.websocket.send_json(message)


@register_streaming_callback("RetrievalQA")
class AsyncRetrievalQAStreamingCallback(AsyncBaseRetrievalQAStreamingCallback):
    """AsyncStreamingResponseCallback handler for RetrievalQA."""

    pass


@register_streaming_callback("VectorDBQA")
class AsyncVectorDBQAStreamingCallback(AsyncBaseRetrievalQAStreamingCallback):
    """AsyncStreamingResponseCallback handler for VectorDBQA."""

    pass


@register_websocket_callback("RetrievalQA")
class AsyncRetrievalQAWebsocketCallback(AsyncBaseRetrievalQAWebsocketCallback):
    """AsyncWebsocketCallback handler for RetrievalQA."""

    pass


@register_websocket_callback("VectorDBQA")
class AsyncVectorDBQAWebsocketCallback(AsyncBaseRetrievalQAWebsocketCallback):
    """AsyncWebsocketCallback handler for VectorDBQA."""

    pass



from abc import abstractmethod
from typing import Any

from fastapi import WebSocket
from langchain.callbacks.base import AsyncCallbackHandler
from pydantic import BaseModel, Field
from starlette.types import Message, Send

from lanarky.schemas import WebsocketResponse


class AsyncLanarkyCallback(AsyncCallbackHandler, BaseModel):
    """Async Callback handler for FastAPI StreamingResponse."""

    @property
    def always_verbose(self) -> bool:
        """Whether to call verbose callbacks even if verbose is False."""
        return True

    class Config:
        arbitrary_types_allowed = True

    @abstractmethod
    def _construct_message(self, message: str) -> Any:  # pragma: no cover
        """Construct a Message from a string."""
        pass


class AsyncStreamingResponseCallback(AsyncLanarkyCallback):
    """Async Callback handler for FastAPI StreamingResponse."""

    send: Send = Field(...)

    def _construct_message(self, message_str: str) -> Message:
        """Construct a Message from a string."""
        return {"type": "http.response.body", "body": message_str, "more_body": True}


class AsyncWebsocketCallback(AsyncLanarkyCallback):
    """Async Callback handler for FastAPI websocket connection."""

    websocket: WebSocket = Field(...)
    response: WebsocketResponse = Field(...)

    def _construct_message(self, message_str: str) -> dict:
        """Construct a WebsocketResponse from a string."""
        return {**self.response.dict(), **{"message": message_str}}



from functools import lru_cache

from pydantic import BaseSettings


class Settings(BaseSettings):
    title: str = "Chatbot Playground"

    # gradio settings
    api_url: str = "http://localhost:8000"
    api_endpoint: str = "/chat"
    gradio_path: str = "/gradio"

    class Config:
        env_file = ".env"


@lru_cache(maxsize=1)
def get_settings() -> Settings:
    return Settings()



from typing import Any, Callable, Dict, List, Optional, Tuple


def register(key: str, _registry: Dict[str, Tuple[Any, List[str]]]) -> Any:
    """Add a class/function to a registry with required keyword arguments.

    ``_registry`` is a dictionary mapping from a key to a tuple of the class/function
    and a list of required keyword arguments, if keyword arguments are passed. Otherwise
    it is a dictionary mapping from a key to the class/function.
    """

    def _register_cls(cls: Any, required_kwargs: Optional[List] = None) -> Any:
        if key in _registry:
            raise KeyError(f"{cls} already registered as {key}")
        _registry[key] = cls if required_kwargs is None else (cls, required_kwargs)
        return cls

    return _register_cls


STREAMING_CALLBACKS: Dict[str, Any] = {}
WEBSOCKET_CALLBACKS: Dict[str, Any] = {}


def register_streaming_callback(key: str) -> Callable:
    """Register an streaming callback handler."""

    def _register_cls(cls: Any) -> Callable:
        register(key, STREAMING_CALLBACKS)(cls=cls)
        return cls

    return _register_cls


def register_websocket_callback(key: str) -> Callable:
    """Register an websocket callback handler."""

    def _register_cls(cls: Any) -> Callable:
        register(key, WEBSOCKET_CALLBACKS)(cls=cls)
        return cls

    return _register_cls
